from typing import Optional

import pytest
from confluent_kafka import KafkaError

from wunderkafka.callbacks import info_callback, error_callback


class Message(object):

    def topic(self) -> str:
        return 'test'

    def partition(self) -> Optional[int]:
        return 1


@pytest.mark.parametrize("callback", [error_callback, info_callback])
@pytest.mark.parametrize("callback_args", [(None, Message()), (KafkaError(10), Message())])
def test_just_print(callback, callback_args) -> None:
    callback(*callback_args)



# -*- coding: utf-8 -*-

import datetime

import pytest

from wunderkafka.hotfixes.watchdog.krb.ticket import clean_stdout, get_datetime

STDOUT1 = """
Ticket cache: FILE:/tmp/krb5cc_1000
Default principal: my.user@MYDOMAIN.COM

Valid starting       Expires              Service principal
05/19/2021 11:39:14  05/20/2021 11:39:14  krbtgt/MYDOMAIN.COM@MYDOMAIN.COM
        renew until 05/26/2021 11:39:14
"""

STDOUT2 = """
klist: No credentials cache found (filename: /tmp/krb5cc_1000)
"""

STDOUT3 = """
Ticket cache: FILE:/tmp/krb5cc_1000
Default principal: service.user@MYDOMAIN.COM

Valid starting     Expires            Service principal
05/18/21 21:02:14  05/19/21 21:02:14  krbtgt/MYDOMAIN.COM@MYDOMAIN.COM
        renew until 05/25/21 21:02:14
05/18/21 21:02:14  05/19/21 07:02:14  kafka/kafka-broker-01.MYDOMAIN.COM@MYDOMAIN.COM
        renew until 05/25/21 21:02:14
05/18/21 21:02:14  05/19/21 07:02:14  HTTP/kafka-broker-01.MYDOMAIN.COM@MYDOMAIN.COM
        renew until 05/25/21 21:02:14
"""


def test_parse_dates():
    unparsable = ('None', 'Expires')
    for st in unparsable:
        assert get_datetime(st) is None

    parsable = {
        '09.04.2021 21:45:41': 4,
        '04/15/2021 20:21:58': 15,
        '15.05.2021 11:22:12': 15,
        '04/09/2021 21:45:41': 9,
        '07/02/21 17:44:03': 2,
    }
    for st, day in parsable.items():
        dt = get_datetime(st)
        assert isinstance(dt, datetime.datetime)
        assert dt.year == 2021
        assert dt.day == day


user = 'my.user'
service_user = 'service.user'
REALM = 'MYDOMAIN.COM'


def test_check_stdout_user():
    res = clean_stdout(STDOUT1, user, REALM)
    assert len(res) == 1

    with pytest.raises(ValueError):
        clean_stdout(STDOUT1, service_user, REALM)

    # FixMe (tribunsky.kir): accidentally, BUG
    # with pytest.raises(ValueError):
    #     clean_stdout(STDOUT1, service_user, 'MYDOMAIN.COM@MYDOMAIN.COM')


def test_no_tickets():
    with pytest.raises(ValueError):
        clean_stdout(STDOUT2, service_user, REALM)

    with pytest.raises(ValueError):
        clean_stdout(STDOUT2, user, REALM)


def test_service_user():
    res = clean_stdout(STDOUT3, service_user, REALM)
    assert len(res) == 2

    with pytest.raises(ValueError):
        clean_stdout(STDOUT3, user, REALM)



import time
import datetime

import pytest

from wunderkafka.consumers.subscription import Offset, Timestamp, choose_offset


def test_nothing_set() -> None:
    assert choose_offset(None, None, None, None) is None


def test_from_beginning() -> None:
    left = choose_offset(from_beginning=True)
    right = choose_offset(from_beginning=False)
    assert left != right
    assert isinstance(left, Offset)
    assert isinstance(right, Offset)


def test_set_ts() -> None:
    now = time.time() * 1000
    res = choose_offset(ts=now)
    assert isinstance(res, Timestamp)
    assert res.value == int(now)

    with pytest.raises(ValueError):
        choose_offset(ts=now, from_beginning=True)


def test_set_timedelta() -> None:
    target = time.time() - 2 * 60
    res = choose_offset(with_timedelta=datetime.timedelta(minutes=2))
    assert isinstance(res, Timestamp)
    assert res.value == pytest.approx(target*1000)

    with pytest.raises(ValueError):
        choose_offset(with_timedelta=datetime.timedelta(minutes=2), from_beginning=True)


def test_set_offset() -> None:
    target = 12345
    res = choose_offset(offset=target)
    assert res.value == target
    assert isinstance(res, Offset)



# -*- coding: utf-8 -*-

import os
import time
import datetime
import subprocess
from typing import Set, Optional

from dateutil import parser

from wunderkafka.logger import logger

# ToDo (tribunsky.kir): move it to platform as it is linux implementation


def check_posix() -> None:
    """
    Check if POSIX locale available.

    Why subprocessing:
    Python's `locale` and linux's `locale -a` behaves differently.

    Example:

    .. code-block:: console

        root@f1cde05f6fdd:/# locale -a
        C
        C.UTF-8
        POSIX
        root@f1cde05f6fdd:/# python
        Python 3.9.7 (default, Sep  3 2021, 20:10:26)
        [GCC 10.2.1 20210110] on linux
        Type "help", "copyright", "credits" or "license" for more information.
        >>> import locale
        >>> len(locale.locale_alias.items())
        588

    :returns:
    """
    try:
        proc = subprocess.run(['locale', '-a'], stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as exc:
        logger.error(exc.output)
        logger.error(exc.stdout)
        logger.error(exc.stderr)
        logger.error('locale exit code: {0}'.format(exc.returncode))
    else:
        lines = {line.strip() for line in proc.stdout.split(b'\n')}
        if b'POSIX' in lines:
            logger.debug('POSIX locale found.')
            return
        if b'posix' in {line.lower() for line in lines}:
            logger.warning('Found POSIX, but in lower-case. Please, check `locale -a` ({0})'.format(lines))
            return
        logger.warning("Couldn't find any POSIX in locales. May misbehave.")


def clean_stdout(stdout: str, krb_user: str = '', krb_realm: str = '') -> Set[str]:
    date_lines = set()
    for line in stdout.split('\n'):
        if krb_realm in line and krb_user not in line:
            _, _, dt, tm, *_ = line.split()
            date_lines.add('{0} {1}'.format(dt, tm))

    if not date_lines:
        raise ValueError('Found no expiration dates. Please, check stdout:\n{0}'.format(stdout))
    logger.debug('Got {0} expiration dates'.format(len(date_lines)))
    return date_lines


def get_expiration_ts(krb_user: str, krb_realm: str, default_timeout: float = 60.0) -> float:
    """
    Get nearest ts when to refresh kerberos ticket, if we can.

    If we can't (no klist, no dates parsed), default timeout to refresh is 60.0 seconds is like in librdkafka.

    If we parsed some invalid data in the past, current timestamp is provided to refresh immediately.
    """
    klist_cmd = ['klist']
    # Using default env, BUT with overridden locale:
    # POSIX locale shall exist in any environment and klist changes it's output
    new_env = {**os.environ, 'LC_ALL': 'POSIX'}
    # logger.debug('Subprocess env to run klist: {0}'.format(new_env))
    try:
        proc = subprocess.run(klist_cmd, stdout=subprocess.PIPE, check=True, env=new_env)
    except subprocess.CalledProcessError as exc:
        logger.error(exc.output)
        logger.error(exc.stdout)
        logger.error(exc.stderr)
        logger.error('klist exit code: {0}'.format(exc.returncode))
        return time.time() + default_timeout
    else:
        expire_dates = []
        for dt_str in clean_stdout(proc.stdout.decode(), krb_user, krb_realm):
            dt = get_datetime(dt_str)
            if dt is not None:
                expire_dates.append(dt)
        logger.debug('Parsed dates: {0}'.format(expire_dates))
        if not expire_dates:
            logger.warning('Got not expiration dates')
            return time.time() + default_timeout
        # expire_dates.append(datetime.datetime.now() + datetime.timedelta(seconds=default_timeout))
        delta = min(expire_dates) - datetime.datetime.now()
        timeout = delta.total_seconds()
        if timeout <= 1:
            return time.time()
        return time.time() + timeout/2.0


def get_datetime(string: str) -> Optional[datetime.datetime]:
    try:
        return parser.parse(string)
    except (ValueError, OverflowError):
        logger.warning('Unable to parse {0}'.format(string))
    return None



import io
from json import loads
from typing import Any, Dict, Optional
from dataclasses import asdict, is_dataclass

from fastavro import parse_schema, schemaless_writer

from wunderkafka.serdes.abc import AbstractSerializer
from wunderkafka.serdes.avro.types import FastAvroParsedSchema


class FastAvroSerializer(AbstractSerializer):
    def __init__(self) -> None:
        self._cache: Dict[str, FastAvroParsedSchema] = {}

    def serialize(self, schema: str, obj: Any, header: Optional[bytes] = None) -> bytes:
        if schema not in self._cache:
            self._cache[schema] = parse_schema(loads(schema))
        writer_schema = self._cache[schema]
        with io.BytesIO() as buffer:
            if header is not None:
                buffer.write(header)
            schemaless_writer(buffer, writer_schema, obj)
            return buffer.getvalue()


class AvroModelSerializer(AbstractSerializer):

    def __init__(self) -> None:
        self._serializer = FastAvroSerializer()

    def serialize(self, schema: str, payload: Any, header: Optional[bytes] = None) -> bytes:
        payload = asdict(payload) if is_dataclass(payload) else dict(payload)
        return self._serializer.serialize(schema, payload, header)



import sys
import json
import time
from typing import Optional
from datetime import datetime
from dataclasses import dataclass

import pytest
from pydantic import Field, BaseModel, BaseSettings

from wunderkafka.compat.types import AvroModel
from wunderkafka.serdes.avromodel import derive


@dataclass
class SomeData(AvroModel):
    field1: int
    field2: str


@dataclass
class SomeDefaultData(AvroModel):
    field1: Optional[int] = None
    field2: Optional[str] = None


class Metrics(BaseModel):
    line_speed: Optional[int]
    defect_detected: Optional[bool] = False
    model_on: Optional[bool] = False
    squad_number: int = 0


class Metric(BaseSettings):
    line_speed: Optional[int]
    defect_detected: Optional[bool] = False
    model_on: Optional[bool] = False
    squad_number: int = 0


class Mixed(BaseModel):
    text: str
    value: float
    integer: int = 0
    string: str = 'str'


# Pydantic allows following of non-default arguments, but dataclasses are not.
class ImageData(BaseModel):
    name: str = 'str'
    image: bytes
    camera: str
    ts: float = Field(default_factory=time.time)


class OtherImageData(BaseModel):
    image: bytes
    camera: str
    ts: float = Field(default_factory=time.time)
    name: str = 'str'


class TsWithMeta(BaseModel):
    ts: datetime = Field(default_factory=datetime.now)

    class Meta:
        namespace = 'com.namespace.my'
        name = 'MsgKey'


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_dataclass() -> None:
    schema = derive(SomeData, topic='test_data_1')
    assert json.loads(schema) == {
        'type': 'record',
        'name': 'test_data_1_value',
        'fields': [
            {
                'name': 'field1',
                'type': 'long',
            },
            {
                'name': 'field2',
                'type': 'string',
            },
        ],
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_dataclass_defaults() -> None:
    schema = derive(SomeDefaultData, topic='test_data_2')

    assert json.loads(schema) == {
        'type': 'record',
        'name': 'test_data_2_value',
        'fields': [
            {
                'name': 'field1',
                'type': ['null', 'long'],
                'default': None,
            },
            {
                'name': 'field2',
                'type': ['null', 'string'],
                'default': None,
            },
        ],
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_pydantic_with_defaults() -> None:
    schema = derive(Metrics, topic='some_topic')

    assert json.loads(schema) == {
      "type": "record",
      "name": "some_topic_value",
      "fields": [
        {
          "name": "line_speed",
          "type": [
            "long",
            "null"
          ]
        },
        {
          "name": "defect_detected",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "model_on",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "squad_number",
          "type": "long",
          "default": 0,
        }
      ]
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_pydantic_defaults() -> None:
    schema = derive(Mixed, topic='topic')

    assert json.loads(schema) == {
        'type': 'record',
        'name': 'topic_value',
        'fields': [
            {
                'type': 'string',
                'name': 'text',
            },
            {
                'type': 'double',
                'name': 'value',
            },
            {
                'type': 'long',
                'name': 'integer',
                'default': 0,
            },
            {
                'type': 'string',
                'name': 'string',
                'default': 'str'
            },
        ],
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_pydantic_mixed_defaults() -> None:
    s1 = derive(ImageData, topic='topic')
    s2 = derive(OtherImageData, topic='topic')
    assert json.loads(s1) == {
        'type': 'record',
        'name': 'topic_value',
        'fields': [
            {
                'name': 'name',
                'type': 'string',
                'default': 'str'
            },
            {
                'name': 'image',
                'type': 'bytes',
            },
            {
                'name': 'camera',
                'type': 'string',
            },
            {
                'name': 'ts',
                'type': 'double',
            },
        ],
    }
    assert json.loads(s1) != json.loads(s2)


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_pydantic_with_meta() -> None:
    schema = derive(TsWithMeta, topic='topic')

    assert json.loads(schema) == {
        'type': 'record',
        'namespace': 'com.namespace.my',
        'name': 'MsgKey',
        'fields': [
            {
                'name': 'ts',
                'type': {
                    'logicalType': 'timestamp-millis',
                    'type': 'long',
                }
            }
        ],
    }


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_pydantic_base_settings_with_defaults() -> None:
    schema = derive(Metric, topic='some_topic')

    assert json.loads(schema) == {
      "type": "record",
      "name": "some_topic_value",
      "fields": [
        {
          "name": "line_speed",
          "type": [
            "long",
            "null"
          ]
        },
        {
          "name": "defect_detected",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "model_on",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "squad_number",
          "type": "long",
          "default": 0,
        }
      ]
    }



"""This module contains some predefined callbacks to interact with librdkafka."""

from typing import List, Optional

from confluent_kafka import Message, KafkaError, TopicPartition

from wunderkafka.logger import logger
from wunderkafka.structures import Timestamp
from wunderkafka.consumers.abc import AbstractConsumer


# FixMe (tribunsky.kir): do not mutate consumer from here, try to closure it in consumer itself.
def reset_partitions(consumer: AbstractConsumer, partitions: List[TopicPartition]) -> None:
    """
    Set specific offset for assignment after subscription.

    Depending on type of subscription, will set offset or timestamp.

    :param consumer:            Consumer, which is subscribes to topics.
    :param partitions:          List of TopicPartitions, which is returned from the underlying library.
    """
    new_offsets = consumer.subscription_offsets
    if new_offsets is None:
        logger.warning(
            '{0}: re-assigned (using auto.offset.reset={1})'.format(consumer, consumer.config.auto_offset_reset),
        )
        return
    by_offset = []
    by_ts = []
    for partition in partitions:
        new_offset = new_offsets[partition.topic]
        if new_offset is None:
            by_offset.append(partition)
        else:
            partition.offset = new_offset.value
            if isinstance(new_offset, Timestamp):
                logger.info('Setting {0}...'.format(new_offset))
                by_ts.append(partition)
            else:
                by_offset.append(partition)
    if by_ts:
        by_ts = consumer.offsets_for_times(by_ts)
    new_ptns = by_ts + by_offset
    consumer.assign(new_ptns)
    logger.info('{0} assigned to {1}'.format(consumer, new_ptns))
    consumer.subscription_offsets = None


def info_callback(err: Optional[KafkaError], msg: Message) -> None:
    """
    Log every message delivery.

    :param err:             Error, if any, thrown from confluent-kafka cimpl.
    :param msg:             Message to be delivered.
    """
    if err is None:
        logger.info('Message delivered to {0} partition: {1}'.format(msg.topic(), msg.partition()))
    else:
        logger.error('Message failed delivery: {0}'.format(err))


def error_callback(err: Optional[KafkaError], _: Message) -> None:
    """
    Log only failed message delivery.

    :param err:             Error, if any, thrown from confluent-kafka cimpl.
    :param _:               Message to be delivered (unused, but needed to not break callback signature).
    """
    if err:
        logger.error('Message failed delivery: {0}'.format(err))



from enum import Enum
from typing import Any, Dict, Union, Optional

import requests_kerberos
from pydantic import AnyHttpUrl, BaseSettings


class HTTPKerberosMutualAuth(Enum):
    REQUIRED = requests_kerberos.REQUIRED
    OPTIONAL = requests_kerberos.OPTIONAL
    DISABLED = requests_kerberos.DISABLED


def remap_fields(dct: Dict[str, Any]) -> Dict[str, Any]:
    return {f_name.replace('_', '.'): f_value for f_name, f_value in dct.items()}


class SRConfig(BaseSettings):
    url: AnyHttpUrl
    ssl_ca_location: Optional[str] = None
    ssl_key_location: Optional[str] = None
    ssl_certificate_location: Optional[str] = None
    basic_auth_user_info: Optional[str] = None

    # ToDo (tribunsky-kir): looks like sasl_username and kerberos over HTTP are cloudera-specific.
    sasl_username: Optional[str]
    # ToDo: (tribunsky-kir): I'd prefer to compose the whole HTTP Kerberos stuff as separate subconfig,
    #                        but it entails writing additional logic for sasl username reuse.
    mutual_auth: Optional[HTTPKerberosMutualAuth] = None

    def dict(self, **kwargs: Any) -> Dict[str, Optional[Union[str, int]]]:
        dct = super().dict(**kwargs)
        return remap_fields(dct)



"""
This module contains the high-level pipeline to produce messages with a nested consumer.

It is intended to be testable enough due to composition of dependencies.

All moving parts should be interchangeable in terms of schema, header and serialization handling
(for further overriding^W extending).
"""

import datetime
from typing import Any, List, Union, Optional

from confluent_kafka import Message, TopicPartition

from wunderkafka.types import HeaderParser
from wunderkafka.logger import logger
from wunderkafka.serdes.abc import AbstractDeserializer
from wunderkafka.structures import SchemaMeta, SchemaDescription
from wunderkafka.consumers.abc import AbstractConsumer, AbstractDeserializingConsumer
from wunderkafka.schema_registry.abc import AbstractSchemaRegistry
from wunderkafka.consumers.subscription import TopicSubscription


class HighLevelDeserializingConsumer(AbstractDeserializingConsumer):
    """Deserializing pipeline implementation of extended consumer."""

    def __init__(
        self,
        consumer: AbstractConsumer,
        headers_handler: HeaderParser,
        schema_registry: AbstractSchemaRegistry,
        deserializer: AbstractDeserializer,
    ):
        """
        Init consumer with specific dependencies.

        :param consumer:            Consumer implementation to receive messages.
        :param headers_handler:     Callable to parse binary headers.
        :param schema_registry:     Schema registry client.
        :param deserializer:        Message deserializer.
        """
        self.consumer = consumer
        self._header_parser = headers_handler
        self._registry = schema_registry
        self._deserializer = deserializer

    def subscribe(  # noqa: D102,WPS211  # docstring inherited from superclass.
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        self.consumer.subscribe(
            topics, from_beginning=from_beginning, offset=offset, ts=ts, with_timedelta=with_timedelta,
        )

    def commit(  # noqa: D102,WPS211  # docstring inherited from superclass.
        self,
        message: Optional[Message] = None,
        offsets: Optional[List[TopicPartition]] = None,
        asynchronous: bool = True,
    ) -> Optional[List[TopicPartition]]:
        if message is None and offsets is not None:
            return self.consumer.commit(offsets=offsets, asynchronous=asynchronous)
        if message is not None and offsets is None:
            return self.consumer.commit(message=message, asynchronous=asynchronous)
        # Default behavior
        return self.consumer.commit(message=message, offsets=offsets, asynchronous=asynchronous)

    def consume(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        ignore_keys: bool = False,
    ) -> List[Message]:
        """
        Consume as many messages as we can for a given timeout and decode them.

        :param timeout:         The maximum time to block waiting for messages. Decoding time doesn't count.
        :param num_messages:    The maximum number of messages to receive from broker.
                                Default is 1000000 which was the allowed maximum for librdkafka 1.2.
        :param ignore_keys:     If True, skip key decoding, key will be set to None. Otherwise decode key as usual.

        :return:                A list of Message objects with decoded value() and key() (possibly empty on timeout).
        """
        msgs = self.consumer.batch_poll(timeout, num_messages)
        return self._decoded(msgs, ignore_keys=ignore_keys)

    def _decoded(
        self,
        msgs: List[Message],
        *,
        ignore_keys: bool = False,
    ) -> Any:
        for msg in msgs:
            # ToDo (tribunsky.kir): move it up to consumer as we may want to handle errors w/o deserialization?
            if msg.error():
                logger.info(msg.error())
            else:
                topic = msg.topic()
                msg.set_value(self._decode(topic, msg.value()))
                if ignore_keys:
                    # ToDo (tribunsky.kir): arguable, maybe it's better to do nothing
                    msg.set_key(None)
                else:
                    msg.set_key(self._decode(topic, msg.key(), is_key=True))
        return msgs

    # Todo (tribunsky.kir): arguable: make different composition (headers, SR & deserializer united cache)
    def _decode(self, topic: str, blob: Optional[bytes], *, is_key: bool = False) -> Any:
        if blob is None:
            return None

        # Header is separate in the sake of customisation, e.g. we don't have SR and put schema directly in message
        parsed_header = self._header_parser(blob)
        schema_meta = SchemaMeta(
            topic=topic,
            is_key=is_key,
            header=parsed_header,
        )
        schema_text = self._registry.get_schema_text(schema_meta)
        schema = SchemaDescription(text=schema_text)
        # performance tradeoff: message may be long and we don't want to:
        # - copy the whole tail
        # - have implicit offset as if we read buffer when extracting header
        # so dealing with implicit offset and the whole binary string
        return self._deserializer.deserialize(schema.text, blob, seek_pos=parsed_header.size)



import sys

import pytest

from wunderkafka.factories import AvroConsumer, AvroProducer, ConsumerConfig, ProducerConfig, AvroModelProducer
from wunderkafka.config.schema_registry import SRConfig


def test_init_avro_consumer(sr_url):
    AvroConsumer(ConsumerConfig(group_id='test', sr=SRConfig(url=sr_url)))


def test_init_avro_producer(sr_url):
    AvroProducer(config=ProducerConfig(sr=SRConfig(url=sr_url)), mapping={})


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
def test_init_avro_model_producer(sr_url):
    AvroModelProducer(config=ProducerConfig(sr=SRConfig(url=sr_url)), mapping={})



import json
from typing import Dict, Union, Optional

from requests import HTTPError

from wunderkafka.errors import SchemaRegistryLookupException
from wunderkafka.logger import logger
from wunderkafka.structures import SRMeta, SchemaMeta, ParsedHeader
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry
from wunderkafka.schema_registry.cache import SimpleCache
from wunderkafka.schema_registry.models import RegisteredSchema

RegistrySchema = Dict[str, Union[str, int, Dict[str, Union[str, int]]]]


def _body(
    schema_text: str,
    references: Optional[dict],
) -> RegistrySchema:
    body = {
        'schema': schema_text,
        'schemaType': 'AVRO',
    }
    if references is not None:
        body['references'] = references
    return body


class ConfluentSRClient(AbstractSchemaRegistry):

    def __init__(self, http_client: AbstractHTTPClient, cache: Optional[SimpleCache] = None) -> None:
        self._client = http_client
        self._cache = SimpleCache() if cache is None else cache

    def get_schema_text(self, meta: SchemaMeta) -> str:
        hdr = meta.header
        schema_text = self._cache.get(hdr)
        if schema_text is None:
            logger.debug("Couldn't find header ({0}) in cache, re-requesting...".format(hdr))
            schema_text = self._choose_schema(hdr, meta.subject)
            self._cache.set(hdr, schema_text)
        return schema_text

    def register_schema(self, topic: str, schema_text: str, *, is_key: bool = True) -> SRMeta:
        uid = (topic, schema_text, is_key)
        meta = self._cache.get(uid)
        if meta is None:
            subject = '{0}{1}'.format(topic, '_key' if is_key else '_value')

            subjects = self._client.make_request('subjects/', query={'subjectPrefix': subject})
            if subject not in subjects:
                schema = self._register_schema(subject, schema_text)
            else:
                # ToDo (aa.perelygin): add check compatibility
                #  (https://docs.confluent.io/platform/current/schema-registry/develop/api.html#sr-api-compatibility)
                schema = self._register_new_version(subject, schema_text)

            if isinstance(schema, int):
                schema = self._get_schema(subject, 'latest')

            meta = SRMeta(
                schema_version=schema.version,
                schema_id=schema.schema_id
            )
            self._cache.set(uid, meta)
        return meta

    def _register_schema(
        self,
        subject: str,
        schema_text: str,
        references: Optional[dict] = None,
        *,
        normalize_schemas: bool = False
    ) -> int:
        response = self._client.make_request(
            'subjects/{0}/versions'.format(subject),
            method='POST',
            body=_body(schema_text, references),
            query={'normalize': normalize_schemas}
        )
        return response['id']

    def _choose_schema(self, hdr: ParsedHeader, subject: str):
        try:
            resp = self._client.make_request('schemas/ids/{}'.format(hdr.schema_id), query={'subject': subject})
            return resp['schema']
        except HTTPError:
            raise SchemaRegistryLookupException("Couldn't find schema for {0}".format(hdr))

    def _get_schema(self, subject: str, version_schema: Union[int, str]) -> RegisteredSchema:
        response = self._client.make_request('subjects/{0}/versions/{1}'.format(subject, version_schema))
        schema = json.loads(response['schema'])

        if isinstance(schema, str):
            ref = None
        else:
            ref = schema.get('references')
            if ref is not None:
                del schema['ref']
        schema = RegisteredSchema(
            schema_id=response['id'],
            schema=schema,
            subject=subject,
            version=response['version'],
            references=ref
        )
        return schema

    def _register_new_version(self, subject: str, schema_text: str) -> int:
        client_schema = json.loads(schema_text)
        versions = self._client.make_request('subjects/{0}/versions'.format(subject))
        latest_schema = None
        for ver in versions:
            reg_schema = self._get_schema(subject, ver)
            if client_schema == reg_schema.schema:
                return reg_schema
            latest_schema = reg_schema
        logger.debug('Schema {} not found'.format(subject))

        ref = None
        if latest_schema and not latest_schema.references:
            ref = {
                'name': latest_schema.subject,
                'subject': latest_schema.subject,
                'version': latest_schema.version
            }
        return self._register_schema(subject, schema_text, ref)



from enum import Enum
from types import MappingProxyType
from typing import Any, Dict, Union, Mapping, TypeVar, Optional

from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.config.generated.fields import COMMON_FIELDS, CONSUMER_FIELDS, PRODUCER_FIELDS
from wunderkafka.config.generated.models import RDConsumerConfig, RDProducerConfig

CONF_CONSUMER_FIELDS = MappingProxyType({
    field_name.replace('.', '_'): field_name for field_name in (*COMMON_FIELDS, *CONSUMER_FIELDS)
})

CONF_PRODUCER_FIELDS = MappingProxyType({
    field_name.replace('.', '_'): field_name for field_name in (*COMMON_FIELDS, *PRODUCER_FIELDS)
})

ConfigValues = Union[str, int, bool, float]


def remap_properties(
    dct: Dict[str, Optional[ConfigValues]],
    mapping: Mapping[str, str],
) -> Dict[str, ConfigValues]:
    new_dct = {}
    for f_name, f_value in dct.items():
        if f_value is not None:
            to_add = f_value
            if isinstance(f_value, Enum):
                to_add = f_value.value
            new_dct[mapping[f_name]] = to_add
    return new_dct


# I don't like mixing SR config and librdkafka config,
# but it's more handful for producer w/o schema (no need to nest config for librdkafka)
class ConsumerConfig(RDConsumerConfig):
    sr: Optional[SRConfig]

    def dict(self, **kwargs: Any) -> Dict[str, ConfigValues]:
        dct = super().dict(**kwargs)
        dct.pop('sr')
        return remap_properties(dct, CONF_CONSUMER_FIELDS)

    # class Config:
    #     allow_mutation = False


class ProducerConfig(RDProducerConfig):
    sr: Optional[SRConfig]

    def dict(self, **kwargs: Any) -> Dict[str, ConfigValues]:
        dct = super().dict(**kwargs)
        dct.pop('sr')
        return remap_properties(dct, CONF_PRODUCER_FIELDS)

    # class Config:
    #     allow_mutation = False


RDKafkaConfig = TypeVar('RDKafkaConfig', ConsumerConfig, ProducerConfig)



import sys

import pytest
from pydantic import Field, BaseModel

from wunderkafka.time import now
from wunderkafka.tests import TestProducer, TestHTTPClient
from wunderkafka.serdes.avro import AvroModelSerializer, ConfluentClouderaHeadersHandler
from wunderkafka.serdes.store import AvroModelRepo
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient
from wunderkafka.producers.constructor import HighLevelSerializingProducer


class Signal(BaseModel):
    ts: int = Field(default_factory=now)
    source: str = 'test'
    type: str = 'string'
    id: str = 'string'
    value: str = 'NA'

    class Meta:
        name = 'SignalsTest'


class DefaultKey(BaseModel):
    ts: int = Field(default_factory=now)

    class Meta:
        namespace = "com.localhost"
        name = "SignalsTestKey"


@pytest.mark.skipif(sys.version_info < (3, 7), reason="requires >= Python3.7")
@pytest.mark.parametrize("schema_description", [Signal, (Signal, DefaultKey)])
def test_avro_producer_moving_parts(sr_root, topic, schema_description) -> None:
    fixed_ts = 1632128298534
    test_producer = TestProducer()
    producer = HighLevelSerializingProducer(
        producer=test_producer,
        schema_registry=ClouderaSRClient(TestHTTPClient(sr_root), SimpleCache()),
        header_packer=ConfluentClouderaHeadersHandler().pack,
        serializer=AvroModelSerializer(),
        store=AvroModelRepo(),
        mapping={topic: schema_description},
    )

    try:
        value_type, key_type = schema_description
    except TypeError:
        key = None
        value_type = schema_description
    else:
        key = key_type(ts=fixed_ts)
    value = value_type(ts=fixed_ts)

    producer.send_message(topic, value, key)

    [message] = test_producer.sent
    if key is None:
        assert message.key is None
    else:
        assert message.key == b'\x01\x00\x00\x00\x00\x00\x00\x06\x9d\x00\x00\x00\x01\xcc\xb8\xeb\xa6\x80_'
    assert message.value == b'\x01\x00\x00\x00\x00\x00\x00\x06\x9c\x00\x00\x00\x01\xcc\xb8\xeb\xa6\x80_\x08test\x0cstring\x0cstring\x04NA'



"""Some compatibility dirt to not fail while importing incompatible functionality."""

import warnings

from wunderkafka.compat.constants import PY36

if PY36 is False:
    # package is for python3.7 and higher.
    from dataclasses_avroschema import AvroModel  # noqa: WPS433
else:
    message = ' '.join([
        'Python 3.6 reaches end of lifespan approximately 2021-12.',
        'Please, consider upgrading to python 3.7 or above to use AvroModel.',
    ])
    warnings.warn(message)

    # I "love" mypy. https://github.com/python/mypy/issues/1153
    class AvroModel(object):  # type: ignore  # noqa: WPS440
        """Stub for 3.6."""

        def __init__(self) -> None:
            """
            Stop execution while trying to use it.

            :raises NotImplementedError:        Python 3.6 reaches end of life soon.
            """
            raise NotImplementedError(message)



import datetime
from typing import Any, List, Union, Optional

from confluent_kafka import KafkaError

from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.consumers.subscription import TopicSubscription


class Message(object):
    def __init__(self, topic: str, value: bytes, key: Optional[bytes] = None, error: Optional[KafkaError] = None):
        self._topic = topic
        self._value = value
        self._key = key
        self._error = error

    def value(self) -> bytes:
        return self._value

    def set_value(self, value: Any) -> None:
        self._value = value

    def key(self) -> Optional[bytes]:
        return self._key

    def set_key(self, key: Any) -> None:
        self._key = key

    def topic(self) -> str:
        return self._topic

    def error(self) -> Optional[KafkaError]:
        return self._error


class TestConsumer(BytesConsumer):
    def __init__(self, msgs: List[Message]) -> None:
        self._msgs = msgs

    def subscribe(
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        pass

    def batch_poll(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> List[Message]:
        return self._msgs



"""This module contains helpers to avoid using try/except ImportError pattern."""

import sys

from typing_extensions import Final

PY36: Final = (sys.version_info.major, sys.version_info.minor) == (3, 6)



from typing import Any, Dict, List, Union, Optional

from wunderkafka.errors import SchemaRegistryLookupException
from wunderkafka.logger import logger
from wunderkafka.structures import SRMeta, SchemaMeta, ParsedHeader
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry
from wunderkafka.schema_registry.cache import SimpleCache, AlwaysEmptyCache


# ToDo (tribunsky.kir): we can cache all schemas at once for every header (key)
def choose_schema(header: ParsedHeader, versions: Dict[str, List[Dict[str, Optional[Union[int, str]]]]]) -> str:
    if header.protocol_id == 0:
        raise NotImplementedError("Didn't check with confluent for now")

    if header.protocol_id == 1:
        for schema in versions['entities']:
            if schema['schemaMetadataId'] == header.meta_id and schema['version'] == header.schema_version:
                # ToDo: use model to describe SchemaRegistry response.
                txt = schema['schemaText']
                if isinstance(txt, str):
                    return txt

    if header.protocol_id in {2, 3}:
        for schema in versions['entities']:
            if schema['id'] == header.schema_id:
                txt = schema['schemaText']
                if isinstance(txt, str):
                    return txt

    raise SchemaRegistryLookupException("Couldn't find schema for {0}".format(header))


class ClouderaSRClient(AbstractSchemaRegistry):

    def __init__(self, http_client: AbstractHTTPClient, cache: Optional[SimpleCache] = None) -> None:
        self._client = http_client
        self._cache = AlwaysEmptyCache() if cache is None else cache

    def _send_request_for_schema(self, meta: SchemaMeta) -> Any:
        subject = meta.subject
        versions = self._cache.get(subject)
        if versions is None:
            versions = self._client.make_request('schemas/{0}/versions'.format(subject))
            self._cache.set(subject, versions)
        return versions

    def _create_meta(self, subject: str) -> int:
        body = {
            'type': 'avro',
            'schemaGroup': 'Kafka',
            'name': subject,
            'description': subject,
            'compatibility': 'BACKWARD',
            'validationLevel': 'ALL',
        }
        return self._client.make_request('schemas', method='POST', body=body)

    def _create_schema(self, subject: str, schema_text: str) -> int:
        body = {
            "description": subject,
            # TypeError: Object of type 'mappingproxy' is not JSON serializable
            # https://github.com/confluentinc/confluent-kafka-python/issues/610
            "schemaText": schema_text,
        }
        query = {'branch': 'MASTER'}
        return self._client.make_request('schemas/{0}/versions'.format(subject), method='POST', body=body, query=query)

    def get_schema_text(self, meta: SchemaMeta) -> str:
        # ToDo (tribunsky.kir): arguably, the best key is BINARY header, not dataclass.
        hdr = meta.header
        schema_text = self._cache.get(hdr)
        if schema_text is None:
            logger.debug("Couldn't find header ({0}) in cache, re-requesting...".format(hdr))
            versions = self._send_request_for_schema(meta)
            schema_text = choose_schema(hdr, versions)
            self._cache.set(hdr, schema_text)
        return schema_text

    # ToDo (tribunsky.kir): here and in producers/constructor:
    #                       - make hypotehtical meta + prop for subject
    #                       - symmetry with get_schema_text
    def register_schema(self, topic: str, schema_text: str, *, is_key: bool = True) -> SRMeta:
        uid = (topic, schema_text, is_key)
        meta = self._cache.get(uid)
        if meta is None:
            suffix = ':k' if is_key else ''
            subject = '{0}{1}'.format(topic, suffix)
            # ToDo (tribunsky.kir): three queris is bad. Either we should know protocol id,
            #                       e.g. to skip last query (we do not need id always)
            meta_id = self._create_meta(subject)
            schema_version = self._create_schema(subject, schema_text)
            versions = self._client.make_request('schemas/{0}/versions'.format(subject))
            for mt in versions['entities']:
                if mt['schemaMetadataId'] == meta_id and mt['version'] == schema_version:
                    meta = SRMeta(
                        meta_id=meta_id,
                        schema_version=schema_version,
                        schema_id=mt['id']
                    )
            self._cache.set(uid, meta)
        return meta

    # def load_all(self):
    #     response = self._client.make_request('schemas')
    #     entities = response['entities']
    #     logger.info('Got {0} schemas'.format(len(entities)))



"""This module contains implementation of extended confluent-kafka Consumer's API."""

import atexit
import datetime
from typing import Dict, List, Union, Optional

from wunderkafka.time import now
from wunderkafka.types import HowToSubscribe
from wunderkafka.config import ConsumerConfig
from wunderkafka.errors import ConsumerException
from wunderkafka.logger import logger
from wunderkafka.callbacks import reset_partitions
from wunderkafka.consumers.abc import Message, AbstractConsumer
from wunderkafka.consumers.subscription import TopicSubscription
from wunderkafka.hotfixes.watchdog.types import Watchdog


class BytesConsumer(AbstractConsumer):
    """Consumer implementation of extended interface for raw messages."""

    # FixMe (tribunsky.kir): add watchdog page reference
    def __init__(self, config: ConsumerConfig, sasl_watchdog: Optional[Watchdog] = None) -> None:
        """
        Init consumer.

        :param config:          Pydantic model with librdkafka consumer's configuration.
        :param sasl_watchdog:   Callable to handle global state of kerberos auth (see Watchdog).
        """
        super().__init__(config.dict())
        self.subscription_offsets: Optional[Dict[str, HowToSubscribe]] = None

        self._config = config
        self._last_poll_ts: int = now()
        self._sasl_watchdog = sasl_watchdog
        # ToDo (tribunsky-kir): make it configurable
        atexit.register(self.close)

    def __str__(self) -> str:
        """
        Get human-readable representation of consumer.

        :return:    string with consumer gid.
        """
        return '{0}:{1}'.format(self.__class__.__name__, self._config.group_id)

    def batch_poll(  # noqa: D102  # inherited from superclass.
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> List[Message]:
        if self._sasl_watchdog is not None:
            self._sasl_watchdog()

        # ToDo (tribunsky.kir): naybe it better to use on_lost callback within subscribe()
        dt = now() - self._last_poll_ts
        if dt > self._config.max_poll_interval_ms:
            msg = 'Exceeded max.poll.interval.ms ({0}): {1}'.format(self._config.max_poll_interval_ms, dt)

            if raise_on_lost:
                # ToDo (tribunsky.kir): resubscribe by ts?
                raise ConsumerException(msg)
            logger.warning(msg)

        msgs = self.consume(num_messages=num_messages, timeout=timeout)
        self._last_poll_ts = now()
        return msgs

    # ToDo (tribunsky.kir): do not override original API and wrap it in superclass
    def subscribe(  # noqa: D102,WPS211  # inherited from superclass.
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        subscriptions = {}
        for tpc in topics:
            if isinstance(tpc, str):
                tpc = TopicSubscription(
                    topic=tpc, from_beginning=from_beginning, offset=offset, ts=ts, with_timedelta=with_timedelta,
                )
            subscriptions[tpc.topic] = tpc.how

        # We have specific subscription at least once
        if any(subscriptions.values()):
            self.subscription_offsets = subscriptions
            # ToDo (tribunsky.kir): avoid mutation of self.subscription_offset and remove it as a field
            super().subscribe(topics=list(self.subscription_offsets), on_assign=reset_partitions)
        else:
            super().subscribe(topics=list(subscriptions))



# I am not gonna to generate single type for every single range of conint/confloat.
# https://github.com/samuelcolvin/pydantic/issues/156

from typing import Callable, Optional

from pydantic import BaseSettings, conint, confloat

# Enums because we can't rely that client code uses linters.
# Of course, it will fail with cimpl.KafkaException, but later, when Consumer/Producer are really initiated
from wunderkafka.config.generated import enums


class RDKafkaConfig(BaseSettings):
    api_version_fallback_ms: conint(ge=0, le=604800000) = 0                                   # type: ignore[valid-type]
    api_version_request: bool = True
    api_version_request_timeout_ms: conint(ge=1, le=300000) = 10000                           # type: ignore[valid-type]
    background_event_cb: Optional[Callable] = None
    bootstrap_servers: Optional[str] = None
    broker_address_family: enums.BrokerAddressFamily = enums.BrokerAddressFamily.any
    broker_address_ttl: conint(ge=0, le=86400000) = 1000                                      # type: ignore[valid-type]
    broker_version_fallback: str = '0.10.0'
    builtin_features: str = ', '.join([
        'gzip',
        'snappy',
        'ssl',
        'sasl',
        'regex',
        'lz4',
        'sasl_gssapi',
        'sasl_plain',
        'sasl_scram',
        'plugins',
        'zstd',
        'sasl_oauthbearer',
    ])
    client_id: str = 'rdkafka'
    client_rack: Optional[str] = None
    closesocket_cb: Optional[Callable] = None
    connect_cb: Optional[Callable] = None
    debug: Optional[str] = None
    default_topic_conf: Optional[Callable] = None
    enable_random_seed: bool = True
    enable_sasl_oauthbearer_unsecure_jwt: bool = False
    enable_ssl_certificate_verification: bool = True
    enabled_events: conint(ge=0, le=2147483647) = 0                                           # type: ignore[valid-type]
    error_cb: Optional[Callable] = None
    interceptors: Optional[Callable] = None
    internal_termination_signal: conint(ge=0, le=128) = 0                                     # type: ignore[valid-type]
    log_cb: Optional[Callable] = None
    log_connection_close: bool = True
    log_level: conint(ge=0, le=7) = 6                                                         # type: ignore[valid-type]
    log_queue: bool = False
    log_thread_name: bool = True
    max_in_flight: conint(ge=1, le=1000000) = 1000000                                         # type: ignore[valid-type]
    max_in_flight_requests_per_connection: conint(ge=1, le=1000000) = 1000000                 # type: ignore[valid-type]
    message_copy_max_bytes: conint(ge=0, le=1000000000) = 65535                               # type: ignore[valid-type]
    message_max_bytes: conint(ge=1000, le=1000000000) = 1000000                               # type: ignore[valid-type]
    metadata_broker_list: Optional[str] = None
    metadata_max_age_ms: conint(ge=1, le=86400000) = 900000                                   # type: ignore[valid-type]
    metadata_request_timeout_ms: conint(ge=10, le=900000) = 60000                             # type: ignore[valid-type]
    oauthbearer_token_refresh_cb: Optional[Callable] = None
    opaque: Optional[Callable] = None
    open_cb: Optional[Callable] = None
    plugin_library_paths: Optional[str] = None
    receive_message_max_bytes: conint(ge=1000, le=2147483647) = 100000000                     # type: ignore[valid-type]
    reconnect_backoff_max_ms: conint(ge=0, le=3600000) = 10000                                # type: ignore[valid-type]
    reconnect_backoff_ms: conint(ge=0, le=3600000) = 100                                      # type: ignore[valid-type]
    sasl_kerberos_keytab: Optional[str] = None
    sasl_kerberos_kinit_cmd: str = 'kinit -R -t "%{sasl.kerberos.keytab}" -k %{sasl.kerberos.principal} || kinit -t "%{sasl.kerberos.keytab}" -k %{sasl.kerberos.principal}'
    sasl_kerberos_min_time_before_relogin: conint(ge=0, le=86400000) = 60000                  # type: ignore[valid-type]
    sasl_kerberos_principal: str = 'kafkaclient'
    sasl_kerberos_service_name: str = 'kafka'
    sasl_mechanism: str = 'GSSAPI'
    sasl_mechanisms: str = 'GSSAPI'
    sasl_oauthbearer_config: Optional[str] = None
    sasl_password: Optional[str] = None
    sasl_username: Optional[str] = None
    security_protocol: enums.SecurityProtocol = enums.SecurityProtocol.plaintext
    socket_cb: Optional[Callable] = None
    socket_keepalive_enable: bool = False
    socket_max_fails: conint(ge=0, le=1000000) = 1                                            # type: ignore[valid-type]
    socket_nagle_disable: bool = False
    socket_receive_buffer_bytes: conint(ge=0, le=100000000) = 0                               # type: ignore[valid-type]
    socket_send_buffer_bytes: conint(ge=0, le=100000000) = 0                                  # type: ignore[valid-type]
    socket_timeout_ms: conint(ge=10, le=300000) = 60000                                       # type: ignore[valid-type]
    ssl_ca: Optional[Callable] = None
    ssl_ca_location: Optional[str] = None
    ssl_certificate: Optional[Callable] = None
    ssl_certificate_location: Optional[str] = None
    ssl_certificate_pem: Optional[str] = None
    ssl_certificate_verify_cb: Optional[Callable] = None
    ssl_cipher_suites: Optional[str] = None
    ssl_crl_location: Optional[str] = None
    ssl_curves_list: Optional[str] = None
    ssl_endpoint_identification_algorithm: enums.SslEndpointIdentificationAlgorithm = enums.SslEndpointIdentificationAlgorithm.none
    ssl_key: Optional[Callable] = None
    ssl_key_location: Optional[str] = None
    ssl_key_password: Optional[str] = None
    ssl_key_pem: Optional[str] = None
    ssl_keystore_location: Optional[str] = None
    ssl_keystore_password: Optional[str] = None
    ssl_sigalgs_list: Optional[str] = None
    statistics_interval_ms: conint(ge=0, le=86400000) = 0                                     # type: ignore[valid-type]
    stats_cb: Optional[Callable] = None
    throttle_cb: Optional[Callable] = None
    topic_blacklist: Optional[str] = None
    topic_metadata_propagation_max_ms: conint(ge=0, le=3600000) = 30000                       # type: ignore[valid-type]
    topic_metadata_refresh_fast_interval_ms: conint(ge=1, le=60000) = 250                     # type: ignore[valid-type]
    topic_metadata_refresh_interval_ms: conint(ge=-1, le=3600000) = 300000                    # type: ignore[valid-type]
    topic_metadata_refresh_sparse: bool = True


class RDConsumerConfig(RDKafkaConfig):
    # FixMe (tribunsky.kir): add to generator
    group_id: str
    allow_auto_create_topics: bool = False
    auto_commit_interval_ms: conint(ge=0, le=86400000) = 5000                                 # type: ignore[valid-type]
    auto_offset_reset: enums.AutoOffsetReset = enums.AutoOffsetReset.latest
    check_crcs: bool = False
    consume_callback_max_messages: conint(ge=0, le=1000000) = 0                               # type: ignore[valid-type]
    consume_cb: Optional[Callable] = None
    coordinator_query_interval_ms: conint(ge=1, le=3600000) = 600000                          # type: ignore[valid-type]
    enable_auto_commit: bool = True
    enable_auto_offset_store: bool = True
    enable_partition_eof: bool = False
    fetch_error_backoff_ms: conint(ge=0, le=300000) = 500                                     # type: ignore[valid-type]
    fetch_max_bytes: conint(ge=0, le=2147483135) = 52428800                                   # type: ignore[valid-type]
    fetch_message_max_bytes: conint(ge=1, le=1000000000) = 1048576                            # type: ignore[valid-type]
    fetch_min_bytes: conint(ge=1, le=100000000) = 1                                           # type: ignore[valid-type]
    fetch_wait_max_ms: conint(ge=0, le=300000) = 500                                          # type: ignore[valid-type]
    group_instance_id: Optional[str] = None
    group_protocol_type: str = 'consumer'
    heartbeat_interval_ms: conint(ge=1, le=3600000) = 3000                                    # type: ignore[valid-type]
    isolation_level: enums.IsolationLevel = enums.IsolationLevel.read_committed
    max_partition_fetch_bytes: conint(ge=1, le=1000000000) = 1048576                          # type: ignore[valid-type]
    max_poll_interval_ms: conint(ge=1, le=86400000) = 300000                                  # type: ignore[valid-type]
    offset_commit_cb: Optional[Callable] = None
    partition_assignment_strategy: str = 'range,roundrobin'
    queued_max_messages_kbytes: conint(ge=1, le=2097151) = 65536                              # type: ignore[valid-type]
    queued_min_messages: conint(ge=1, le=10000000) = 100000                                   # type: ignore[valid-type]
    rebalance_cb: Optional[Callable] = None
    session_timeout_ms: conint(ge=1, le=3600000) = 10000                                      # type: ignore[valid-type]


class RDProducerConfig(RDKafkaConfig):
    acks: conint(ge=-1, le=1000) = -1                                                         # type: ignore[valid-type]
    batch_num_messages: conint(ge=1, le=1000000) = 10000                                      # type: ignore[valid-type]
    batch_size: conint(ge=1, le=2147483647) = 1000000                                         # type: ignore[valid-type]
    compression_codec: enums.CompressionCodec = enums.CompressionCodec.none
    compression_level: conint(ge=-1, le=12) = -1                                              # type: ignore[valid-type]
    compression_type: enums.CompressionType = enums.CompressionType.none
    delivery_report_only_error: bool = False
    delivery_timeout_ms: conint(ge=0, le=2147483647) = 300000                                 # type: ignore[valid-type]
    dr_cb: Optional[Callable] = None
    dr_msg_cb: Optional[Callable] = None
    enable_gapless_guarantee: bool = False
    enable_idempotence: bool = False
    linger_ms: confloat(ge=0, le=900000) = 5.0                                                # type: ignore[valid-type]
    message_send_max_retries: conint(ge=0, le=10000000) = 2                                   # type: ignore[valid-type]
    message_timeout_ms: conint(ge=0, le=2147483647) = 300000                                  # type: ignore[valid-type]
    msg_order_cmp: Optional[Callable] = None
    partitioner: str = 'consistent_random'
    partitioner_cb: Optional[Callable] = None
    queue_buffering_backpressure_threshold: conint(ge=1, le=1000000) = 1                      # type: ignore[valid-type]
    queue_buffering_max_kbytes: conint(ge=1, le=2147483647) = 1048576                         # type: ignore[valid-type]
    queue_buffering_max_messages: conint(ge=1, le=10000000) = 100000                          # type: ignore[valid-type]
    queue_buffering_max_ms: confloat(ge=0, le=900000) = 5.0                                   # type: ignore[valid-type]
    queuing_strategy: enums.QueuingStrategy = enums.QueuingStrategy.fifo
    request_required_acks: conint(ge=-1, le=1000) = -1                                        # type: ignore[valid-type]
    request_timeout_ms: conint(ge=1, le=900000) = 5000                                        # type: ignore[valid-type]
    retries: conint(ge=0, le=10000000) = 2                                                    # type: ignore[valid-type]
    retry_backoff_ms: conint(ge=1, le=300000) = 100                                           # type: ignore[valid-type]
    transaction_timeout_ms: conint(ge=1000, le=2147483647) = 60000                            # type: ignore[valid-type]
    transactional_id: Optional[str] = None



"""
This module contains the high-level pipeline to produce messages with a nested producer.

It is intended to be testable enough due to composition of dependencies.

All moving parts should be interchangeable in terms of schema, header and serialization handling
(for further overriding^W extending).
"""

from typing import Any, Dict, Tuple, Optional

from wunderkafka.types import MsgKey, MsgValue, TopicName, HeaderPacker, DeliveryCallback, MessageDescription
from wunderkafka.logger import logger
from wunderkafka.callbacks import error_callback
from wunderkafka.serdes.abc import AbstractSerializer, AbstractDescriptionStore
from wunderkafka.structures import SRMeta, SchemaDescription
from wunderkafka.producers.abc import AbstractProducer, AbstractSerializingProducer
from wunderkafka.schema_registry.abc import AbstractSchemaRegistry


class HighLevelSerializingProducer(AbstractSerializingProducer):
    """Serializing pipeline implementation of extended producer."""

    def __init__(  # noqa: WPS211  # ToDo (tribunsky.kir): rethink building of producer.
        #                                                  Maybe single Callable (like in kafka-python) much better.
        self,
        producer: AbstractProducer,
        schema_registry: AbstractSchemaRegistry,
        header_packer: HeaderPacker,
        serializer: AbstractSerializer,
        store: AbstractDescriptionStore,
        # ToDo: switch mapping to something like consumer's TopicSubscription?
        mapping: Optional[Dict[TopicName, MessageDescription]] = None,
        *,
        protocol_id: int = 1,
        lazy: bool = False,
    ) -> None:
        """
        Init producer with specific dependencies and prepare it to work against specified topic(s).

        :param producer:            Producer implementation to send messages.
        :param schema_registry:     Schema registry client.
        :param header_packer:       Callable to form binary headers.
        :param serializer:          Message serializer.
        :param store:               Specific store to provide schema text extraction from schema description.
        :param mapping:             Per-topic definition of value and/or key schema descriptions.
        :param protocol_id:         Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)
        :param lazy:                If True, defer schema registry publication, otherwise schema will be registered
                                    before the first message sending.
        """
        self._mapping = mapping or {}

        # ToDo (tribunsky.kir): look like wrong place. Maybe it's better to highlight an entity
        #                       of Schema which may be checked or not. Then input mapping may be eliminated (or not).
        self._checked: Dict[Tuple[str, str, bool], SRMeta] = {}

        self._store = store
        self._sr = schema_registry
        self._serializer = serializer
        self._producer = producer
        self._header_packer = header_packer
        self._protocol_id = protocol_id

        for topic, description in self._mapping.items():
            if isinstance(description, (tuple, list)):
                msg_value, msg_key = description
                self.set_target_topic(topic, msg_value, msg_key, lazy=lazy)
            else:
                self.set_target_topic(topic, description, lazy=lazy)

    def flush(self, timeout: Optional[float] = None) -> int:   # noqa: D102  # docstring inherited from superclass.
        if timeout is None:
            return self._producer.flush()
        return self._producer.flush(timeout)

    def set_target_topic(  # noqa: D102  # docstring inherited from superclass.
        self,
        topic: str,
        value: Any,  # noqa: WPS110  # Domain. inherited from superclass.
        key: Any = None,
        *,
        lazy: bool = False,
    ) -> None:
        self._store.add(topic, value, key)
        if not lazy:
            value_descr = self._store.get(topic)
            self._check_schema(topic, value_descr)
            if key is not None:
                key_descr = self._store.get(topic, is_key=True)
                self._check_schema(topic, key_descr, is_key=True)

    def send_message(  # noqa: D102,WPS211  # inherited from superclass.
        self,
        topic: str,
        value: MsgValue = None,  # noqa: WPS110  # Domain. inherited from superclass.
        key: MsgKey = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = error_callback,
        *args: Any,
        blocking: bool = False,
        protocol_id: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        if protocol_id is None:
            protocol_id = self._protocol_id
        encoded_value = self._encode(topic, value, protocol_id)
        encoded_key = self._encode(topic, key, protocol_id, is_key=True)

        self._producer.send_message(
            topic,
            encoded_value,
            encoded_key,
            partition,
            on_delivery,
            blocking=blocking,
            *args,
            **kwargs,
        )

    def _check_schema(
        self,
        topic: str,
        schema: Optional[SchemaDescription],
        *,
        is_key: bool = False,
        force: bool = False,
    ) -> SRMeta:
        """
        Ensure that we have schema's ids necessary to create message's header.

        :param topic:   Target topic against which we are working for this call.
        :param schema:  Schema container to be registered. Should contain text.
        :param is_key:  If True, schema will be requested as message key, message value otherwise.
        :param force:   If True, do not reuse cached results, always request Schema Registry.

        :raises ValueError: If received no schema from DescriptionStore.

        :return:        Container with schema's ids.
        """
        if schema is None:
            raise ValueError("Couldn't check schema from store.")
        uid = (topic, schema.text, is_key)

        if force is False:
            meta = self._checked.get(uid)
            if meta is not None:
                return meta
        meta = self._sr.register_schema(topic, schema.text, is_key=is_key)
        self._checked[uid] = meta
        return meta

    # ToDo (tribunsky.kir): Maybe, this method should be public for more simple extension.
    #                       'Template' pattern never works anyway.
    def _encode(self, topic: str, obj: Any, protocol_id: int, is_key: bool = False) -> Optional[bytes]:  # noqa: WPS110
        if obj is None:
            return None

        schema = self._store.get(topic, is_key=is_key)
        if schema is not None:
            # ToDo (tribunsky.kir): make it symmetrical with HeaderParser
            available_meta = self._check_schema(topic, schema, is_key=is_key)
            # ToDo (tribunsky.kir): add to (de)serialization ability to include the whole schema to the header.
            header = self._header_packer(protocol_id, available_meta)
            return self._serializer.serialize(schema.text, obj, header)
        # ToDo (tribunsky.kir): In the sake of mypy, re-do (add strict flag or something like that).
        logger.warning('Missing schema for {0} (key: {1}'.format(topic, is_key))
        return None



import pytest
from pydantic import ValidationError

from wunderkafka import SRConfig, BytesConsumer, BytesProducer, ConsumerConfig, ProducerConfig


@pytest.fixture
def boostrap_servers() -> str:
    return 'localhost:9093'


def test_init_consumer(boostrap_servers) -> None:
    config = ConsumerConfig(group_id='my_group', bootstrap_servers=boostrap_servers)
    consumer = BytesConsumer(config)
    print(consumer)

    with pytest.raises(AttributeError):
        consumer.config = ConsumerConfig(group_id='my_other_group', bootstrap_servers=boostrap_servers)
    assert consumer.config == config

    consumer.close()


def test_init_producer(boostrap_servers) -> None:
    config = ProducerConfig(bootstrap_servers=boostrap_servers)
    BytesProducer(config)


def test_sr_required_url() -> None:
    with pytest.raises(ValidationError):
        SRConfig()


def test_group_id_required() -> None:
    with pytest.raises(ValidationError):
        ConsumerConfig()



from typing import Tuple, Optional
from dataclasses import dataclass


@dataclass(frozen=True)
class RegisteredSchema:
    schema_id: int
    schema: dict
    subject: str
    version: int

    references: Optional[dict] = None

    @property
    def uniq_key(self) -> Tuple[str, int]:
        return self.subject, self.schema_id



import json
from typing import Any, Union
from pathlib import Path

from wunderkafka.schema_registry.abc import AbstractHTTPClient


class TestHTTPClient(AbstractHTTPClient):

    def __init__(self, root_dir: Union[str, Path]) -> None:
        self._root = Path(root_dir)

    def make_request(self, relative_url: str, method: str = 'GET', body: Any = None, query: Any = None) -> Any:
        filename = self._root / method / '{0}.json'.format(relative_url)
        if relative_url == 'schemas':
            filename = self._root / method / relative_url / body['description'] / '{0}.json'.format(relative_url)
        with open(filename) as fl:
            return json.load(fl)



import io
from json import loads
from typing import Any, Dict, Optional

from fastavro import parse_schema, schemaless_reader

from wunderkafka.serdes.abc import AbstractDeserializer
from wunderkafka.serdes.avro.types import FastAvroParsedSchema


class FastAvroDeserializer(AbstractDeserializer):
    # ToDo (tribunsky.kir): it's better to cache loaded Schema earlier, but then it breaks layers.

    def __init__(self) -> None:
        self._cache: Dict[str, FastAvroParsedSchema] = {}

    def deserialize(self, schema: str, blob: bytes, seek_pos: Optional[int] = None) -> Any:
        if schema not in self._cache:
            self._cache[schema] = parse_schema(loads(schema))
        reader_schema = self._cache[schema]
        with io.BytesIO(blob) as buffer:
            if seek_pos is not None:
                buffer.seek(seek_pos)
            return schemaless_reader(buffer, reader_schema, None, False)



from dataclasses import dataclass


class Mask(object):
    def __init__(self, value: str):
        self._value = value
        self.__unpack = '>{0}'.format(self._value)
        self.__pack = '>b{0}'.format(self._value)

    @property
    def unpack(self) -> str:
        return self.__unpack

    @property
    def pack(self) -> str:
        return self.__pack


@dataclass(frozen=True)
class Protocol(object):
    header_size: int
    mask: Mask



from wunderkafka.hotfixes.watchdog import parse_kinit

USER = 'my.user'
REALM = 'MYDOMAIN.COM'


def test_relative_regular():
    keytab = '{0}.keytab'.format(USER)
    kinit_cmd = 'kinit {0}@{1} -k -t {2}'.format(USER, REALM, keytab)
    assert parse_kinit(kinit_cmd) == (USER, REALM, keytab)


def test_absolute_with_domain():
    keytab = '/home/{0}@{1}/{0}.keytab'.format(USER, REALM.lower())
    kinit_cmd = 'kinit {0}@{1} -k -t {2}'.format(USER, REALM, keytab)
    assert parse_kinit(kinit_cmd) == (USER, REALM, keytab)



"""This module contains custom exceptions for the package."""

from confluent_kafka import KafkaException


class ConsumerException(KafkaException):
    """Break control flow if there are any problems with consumer."""


class SerDesException(RuntimeError):
    """Break control flow if there are any problems with (de)serialization."""


class SerializerException(SerDesException):
    """Break control flow if there are any problems with serialization."""


class DeserializerException(SerDesException):
    """Break control flow if there are any problems with deserialization."""


class SchemaRegistryLookupException(RuntimeError):
    """Break control flow if there are any problems with schema id handling.."""



from typing import Type, Union, Optional
from pathlib import Path

from wunderkafka.types import TopicName, KeySchemaDescription, ValueSchemaDescription
from wunderkafka.serdes.abc import AbstractDescriptionStore
from wunderkafka.compat.types import AvroModel
from wunderkafka.compat.constants import PY36
from wunderkafka.serdes.avromodel import derive


class SchemaTextRepo(AbstractDescriptionStore):

    def add(self, topic: TopicName, value: str, key: str) -> None:
        self._values[topic] = ValueSchemaDescription(text=value)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=key)


def _load_from_file(filename: Path) -> str:
    with open(filename) as fl:
        return fl.read()


# ToDo (tribunsky.kir): refactor it, maybe add hooks to parent class.
#                       Barbara, forgive us. Looks like AbstractDescriptionStore should be generic.
class SchemaFSRepo(AbstractDescriptionStore):

    def add(self, topic: TopicName, value: Union[str, Path], key: Union[str, Path]) -> None:
        self._values[topic] = ValueSchemaDescription(text=_load_from_file(Path(value)))
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=_load_from_file(Path(key)))


class AvroModelRepo(AbstractDescriptionStore):

    def __init__(self) -> None:
        super().__init__()
        if PY36:
            AvroModel()

    # ToDo (tribunsky.kir): change Type[AvroModel] to more general alias + check derivation from python built-ins
    def add(self, topic: TopicName, value: Type[AvroModel], key: Optional[Type[AvroModel]]) -> None:
        self._values[topic] = ValueSchemaDescription(text=derive(value, topic))
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=derive(key, topic, is_key=True))



import struct
from types import MappingProxyType

from wunderkafka.errors import DeserializerException
from wunderkafka.serdes.abc import AbstractProtocolHandler
from wunderkafka.structures import SRMeta, SchemaMeta, ParsedHeader
from wunderkafka.serdes.avro.structures import Mask, Protocol

# see: https://git.io/JvYyC
PROTOCOLS = MappingProxyType({
    # public static final byte CONFLUENT_VERSION_PROTOCOL = 0x0;
    0: Protocol(4, Mask('I')),
    # public static final byte METADATA_ID_VERSION_PROTOCOL = 0x1;
    1: Protocol(12, Mask('qi')),
    # public static final byte VERSION_ID_AS_LONG_PROTOCOL = 0x2;
    2: Protocol(8, Mask('q')),
    # public static final byte VERSION_ID_AS_INT_PROTOCOL = 0x3;
    3: Protocol(4, Mask('I')),
})

MIN_HEADER_SIZE = min([pr.header_size for pr in PROTOCOLS.values()]) + 1


class ConfluentClouderaHeadersHandler(AbstractProtocolHandler):

    def parse(self, blob: bytes) -> ParsedHeader:
        if len(blob) < MIN_HEADER_SIZE:
            raise DeserializerException("Message is too small to decode: ({!r})".format(blob))

        # 1st byte is magic byte.
        [protocol_id] = struct.unpack('>b', blob[0:1])

        protocol = PROTOCOLS.get(protocol_id)
        if protocol is None:
            raise DeserializerException('Unknown protocol extracted from message: {0}'.format(protocol_id))

        # already read 1 byte from header as protocol id
        meta = struct.unpack(protocol.mask.unpack, blob[1:1+protocol.header_size])

        if protocol_id == 1:
            schema_id = None
            schema_meta_id, schema_version = meta
        else:
            [schema_id] = meta
            schema_meta_id = None
            schema_version = None

        return ParsedHeader(
            protocol_id=protocol_id,
            meta_id=schema_meta_id,
            schema_id=schema_id,
            schema_version=schema_version,
            size=protocol.header_size + 1,
        )

    def pack(self, protocol_id: int, meta: SRMeta) -> bytes:
        protocol = PROTOCOLS.get(protocol_id)
        if protocol is None:
            raise ValueError('Unknown protocol_id {0}!'.format(protocol_id))

        if protocol_id == 1:
            if meta.meta_id is None:
                err_msg = 'No meta id for protocol_id {0}. Please, check response from Schema Registry.'.format(
                    protocol_id)
                raise ValueError(err_msg)
            return struct.pack(protocol.mask.pack, protocol_id, meta.meta_id, meta.schema_version)
        else:
            return struct.pack(protocol.mask.pack, protocol_id, meta.schema_id)



"""
Module contains interface-like skeletons for producer.

- inherited from confluent-kafka Producer to use it as nested entity
- high-level producer which is able to handle message's schemas
"""


from abc import ABC, abstractmethod
from typing import Any, Union, Optional

from confluent_kafka import Producer

from wunderkafka.types import MsgKey, MsgValue, DeliveryCallback


class AbstractProducer(Producer):
    """Extension point for the original Producer API."""

    def send_message(  # noqa: WPS211  # ToDo (tribunsky.kir): rethink API?
        self,
        topic: str,
        value: Optional[Union[str, bytes]] = None,  # noqa: WPS110  # Domain.
        key: Optional[Union[str, bytes]] = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = None,
        *args: Any,
        blocking: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Send encoded message to Kafka almost immediately.

        This method overlaps original producers' method.

        :param topic:           Target topic against which we are working for this call.
        :param value:           Message's value encoded object to be sent.
        :param key:             Message's key encoded object to be sent.
        :param partition:       Target partition to produce to. If none, uses configured built-in
                                partitioner by default.
        :param on_delivery:     Callback to be executed on successful/failed delivery.
        :param args:            Other positional arguments to call with original produce() API.
        :param blocking:        If True, will block execution until delivery result retrieved
                                (calls flush() under the hood). Otherwise, will produce message to
                                built-in queue and call poll() to trigger producer's events.
        :param kwargs:          Other keyword arguments to call with original produce() API.


        ..  # noqa: DAR401
        """
        raise NotImplementedError


class AbstractSerializingProducer(ABC):
    """High-level interface for extended producer."""

    @abstractmethod
    def send_message(  # noqa: WPS211  # overlaps API from AbstractProducer, maybe should be mixin.
        self,
        topic: str,
        value: MsgValue = None,  # noqa: WPS110  # Domain, overlaps API from AbstractProducer
        key: MsgKey = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = None,
        *args: Any,
        blocking: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Encode and send message to Kafka.

        This method overlaps AbstractProducer's method and intended to use produce() of real nested producer.

        :param topic:           Target topic against which we are working for this call.
        :param value:           Message's value object to be encoded and sent.
        :param key:             Message's key object to be encoded and sent.
        :param partition:       Target partition to produce to. If none, uses configured built-in
                                partitioner by default.
        :param on_delivery:     Callback to be executed on successful/failed delivery.
        :param args:            Other positional arguments to call with original produce() API.
        :param blocking:        If True, will block execution until delivery result retrieved
                                (calls flush() under the hood). Otherwise, will produce message to
                                built-in queue and call poll() to trigger producer's events.
        :param kwargs:          Other keyword arguments to call with original produce() API.
        """

    # ToDo (tribunsky.kir): change naming. It is like subscribe, but publish/add_topic whatever
    @abstractmethod
    def set_target_topic(self, topic: str, value: Any, key: Any = None, *, lazy: bool = False) -> None:  # noqa: WPS110
        """
        Make producer aware how it should work with specific topic.

        This method is used to instantiate producer with mapping, but also add schemas in runtime.

        :param topic:           Target topic to specify schema description against.
        :param value:           Message's value schema description.
        :param key:             Message's key schema description.
        :param lazy:            If True, do not register schema in registry during __init__,
                                or before the first attempt to send message for a given topic.
                                Otherwise, register schema immediately.
        """

    @abstractmethod
    def flush(self, timeout: float) -> int:
        """
        Wait for all messages in the Producer queue to be delivered.

        This method overlaps original producers' method and will use the nested producer.

        :param timeout:         Maximum time to block.
        :return:                Number of messages still in queue.
        """



from enum import Enum


class BrokerAddressFamily(str, Enum):
    any = 'any'
    v4 = 'v4'
    v6 = 'v6'


class SecurityProtocol(str, Enum):
    plaintext = 'plaintext'
    ssl = 'ssl'
    sasl_plaintext = 'sasl_plaintext'
    sasl_ssl = 'sasl_ssl'


class SslEndpointIdentificationAlgorithm(str, Enum):
    none = 'none'
    https = 'https'


class IsolationLevel(str, Enum):
    read_uncommitted = 'read_uncommitted'
    read_committed = 'read_committed'


class AutoOffsetReset(str, Enum):
    smallest = 'smallest'
    earliest = 'earliest'
    beginning = 'beginning'
    largest = 'largest'
    latest = 'latest'
    end = 'end'
    error = 'error'


class CompressionCodec(str, Enum):
    none = 'none'
    gzip = 'gzip'
    snappy = 'snappy'
    lz4 = 'lz4'
    zstd = 'zstd'


class CompressionType(str, Enum):
    none = 'none'
    gzip = 'gzip'
    snappy = 'snappy'
    lz4 = 'lz4'
    zstd = 'zstd'


class QueuingStrategy(str, Enum):
    fifo = 'fifo'
    lifo = 'lifo'



import json
from typing import Any, Dict
from pathlib import Path

import pytest

from wunderkafka.serdes.store import SchemaFSRepo


@pytest.fixture
def repo() -> SchemaFSRepo:
    return SchemaFSRepo()


@pytest.fixture
def fs_root_dir(fixtures_root: Path) -> Path:
    return fixtures_root / 'fs_repo'


@pytest.fixture
def answer_json() -> Dict[str, Any]:
    return {
     "type": "record",
     "name": "exampleSchema",
     "fields": [
      {
       "name": "name",
       "type": "string",
       "default": ""
      },
      {
       "name": "value",
       "type": "string",
       "default": ""
      }
     ]
    }


@pytest.mark.parametrize('key_schema_path,value_schema_path', [('key.avsc', 'value.avsc')])
def test_fs_repo(
    repo: SchemaFSRepo,
    topic: str,
    key_schema_path: str,
    value_schema_path: str,
    fs_root_dir: Path,
    answer_json: Dict[str, Any]
) -> None:
    repo.add(topic, fs_root_dir / value_schema_path, fs_root_dir / key_schema_path)
    assert repo.get(topic, is_key=True).text == '"long"'
    assert json.loads(repo.get(topic).text) == answer_json



from typing import Any, Dict
from dataclasses import dataclass

import pytest

from wunderkafka.tests import TestConsumer, TestHTTPClient
from wunderkafka.serdes.avro import FastAvroDeserializer, ConfluentClouderaHeadersHandler
from wunderkafka.tests.consumer import Message
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer


@dataclass
class Msg(object):
    payload: bytes
    deserialized: Dict[str, Any]

    def serialized(self, header: bytes) -> bytes:
        return header + self.payload


SIGNAL_MESSAGE = Msg(
    payload=b'\x08test\x0cstring\x0cstring\x04NA\xcc\xb8\xeb\xa6\x80_',
    deserialized={
        'source': 'test',
        'type': 'string',
        'id': 'string',
        'value': 'NA',
        'ts': 1632128298534,
    },
)


HEADERS = (
    b'\x01\x00\x00\x00\x00\x00\x00\x06\x9c\x00\x00\x00\x01',
    b'\x02\x00\x00\x00\x00\x00\x00\x08<',
    b'\x03\x00\x00\x08<',
)


@pytest.mark.parametrize("header", list(HEADERS))
def test_consume_moving_parts(sr_root, topic, header) -> None:
    consumer = HighLevelDeserializingConsumer(
        consumer=TestConsumer([Message(topic, value=SIGNAL_MESSAGE.serialized(header))]),
        schema_registry=ClouderaSRClient(TestHTTPClient(sr_root), SimpleCache()),
        headers_handler=ConfluentClouderaHeadersHandler().parse,
        deserializer=FastAvroDeserializer(),
    )

    consumer.subscribe([topic], from_beginning=True)

    msgs = consumer.consume()
    [message] = msgs
    assert message.key() is None
    assert message.value() == SIGNAL_MESSAGE.deserialized
